#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of segment-reshape-qgis-plugin.
#
#  segment-reshape-qgis-plugin is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  segment-reshape-qgis-plugin is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with segment-reshape-qgis-plugin. If not, see <https://www.gnu.org/licenses/>.

import logging
from contextlib import contextmanager
from enum import Enum, IntEnum
from typing import (
    TYPE_CHECKING,
    Generator,
    Iterator,
    Optional,
    Tuple,
    Union,
    cast,
    overload,
)

from qgis.core import (
    QgsApplication,
    QgsGeometry,
    QgsLineString,
    QgsPoint,
    QgsPointLocator,
    QgsPointXY,
    QgsVectorLayer,
    QgsWkbTypes,
)
from qgis.gui import (
    QgisInterface,
    QgsAbstractMapToolHandler,
    QgsMapCanvas,
    QgsMapMouseEvent,
    QgsMapToolCapture,
    QgsMapToolIdentify,
    QgsRubberBand,
)
from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtGui import QColor, QKeyEvent
from qgis.PyQt.QtWidgets import QApplication
from qgis.utils import iface as iface_
from qgis_plugin_tools.tools.decorations import log_if_fails
from qgis_plugin_tools.tools.i18n import tr
from qgis_plugin_tools.tools.messages import MsgBar

from segment_reshape.geometry import reshape
from segment_reshape.topology import find_related

if TYPE_CHECKING:
    from qgis.core import QgsMapLayer
    from qgis.PyQt.QtWidgets import QAction

iface = cast(QgisInterface, iface_)

LOGGER = logging.getLogger(__name__)


class CoordinateTransformationError(Exception):
    ...


@contextmanager
def _optional_identify_tool(
    tool: Optional[QgsMapToolIdentify] = None,
) -> Iterator[QgsMapToolIdentify]:
    if tool:
        yield tool
    else:
        tool = QgsMapToolIdentify(iface.mapCanvas())
        try:
            yield tool
        finally:
            tool.deleteLater()


@contextmanager
def override_cursor(cursor: Qt.CursorShape) -> Generator[None, None, None]:
    QApplication.setOverrideCursor(cursor)
    try:
        yield
    finally:
        QApplication.restoreOverrideCursor()


COLOR_BLUE = QColor(10, 20, 150)
COLOR_GREY = QColor(211, 211, 211)


class ToolMode(Enum):
    PICK_SEGMENT = "pick_segment"
    RESHAPE = "reshape"


class AddVertexReturn(IntEnum):
    Success = 0
    TransformationError = 2


class SegmentReshapeToolHandler(QgsAbstractMapToolHandler):
    def __init__(self, tool: "SegmentReshapeTool", action: "QAction") -> None:
        super().__init__(tool, action)

    def isCompatibleWithLayer(  # noqa: N802
        self, layer: "QgsMapLayer", context: QgsAbstractMapToolHandler.Context
    ) -> bool:
        return isinstance(layer, QgsVectorLayer) and layer.geometryType() in (
            QgsWkbTypes.GeometryType.PolygonGeometry,
            QgsWkbTypes.GeometryType.LineGeometry,
        )


class SegmentReshapeTool(QgsMapToolCapture):
    _tool_mode: ToolMode

    def __init__(self, canvas: QgsMapCanvas) -> None:
        super().__init__(
            canvas, iface.cadDockWidget(), QgsMapToolCapture.CaptureMode.CaptureLine
        )

        self.old_segment_rubber_band = QgsRubberBand(canvas)
        self.old_segment_rubber_band.setStrokeColor(COLOR_BLUE)
        self.old_segment_rubber_band.setWidth(10)

        self.start_point_indicator_rubber_band = QgsRubberBand(canvas)
        self.start_point_indicator_rubber_band.setStrokeColor(COLOR_GREY)
        self.start_point_indicator_rubber_band.setLineStyle(Qt.PenStyle.DotLine)

        self.find_segment_results = find_related.CommonGeometriesResult(None, [], [])

        self._identify_tool = QgsMapToolIdentify(canvas)

    def activate(self) -> None:
        self._change_to_pick_location_mode()
        super().activate()

    def deactivate(self) -> None:
        self.old_segment_rubber_band.hide()
        self.start_point_indicator_rubber_band.hide()

        return super().deactivate()

    def _change_to_pick_location_mode(self) -> None:
        self._tool_mode = ToolMode.PICK_SEGMENT
        self.setCursor(Qt.CrossCursor)
        self.setAutoSnapEnabled(False)
        self.setAdvancedDigitizingAllowed(False)

        self.old_segment_rubber_band.reset()
        self.start_point_indicator_rubber_band.reset()

        self.find_segment_results = find_related.CommonGeometriesResult(None, [], [])

    def _change_to_reshape_mode_for_geom(
        self, old_geom: QgsGeometry, layer: Optional[QgsVectorLayer] = None
    ) -> None:
        self._tool_mode = ToolMode.RESHAPE
        self.setCursor(
            QgsApplication.getThemeCursor(QgsApplication.Cursor.CapturePoint)
        )
        self.setAutoSnapEnabled(True)
        self.setAdvancedDigitizingAllowed(True)

        self._common_segment_to_reshape = old_geom
        self._active_layer = layer

        self.old_segment_rubber_band.setToGeometry(old_geom, layer)

        self.start_point_indicator_rubber_band.reset()
        self.start_point_indicator_rubber_band.show()
        start_point = self.old_segment_rubber_band.getPoint(0, 0)
        self.start_point_indicator_rubber_band.addPoint(start_point, True)

    def keyPressEvent(self, key_event: QKeyEvent) -> None:  # noqa: N802
        point_count = self.size()
        super().keyPressEvent(key_event)  # handle normal undo and cancel procedures
        if self._tool_mode == ToolMode.RESHAPE:
            if key_event.key() == Qt.Key_Escape:
                self._change_to_pick_location_mode()
            elif (
                key_event.key() in (Qt.Key_Delete, Qt.Key_Backspace)
                and point_count == 1
            ):
                # self.undo() resists to undo the first point so force back to initial
                # reshape mode
                self.stopCapturing()
                self._change_to_reshape_mode_for_geom(
                    self._common_segment_to_reshape, self._active_layer
                )

    @overload
    def addVertex(self, point: QgsPointXY) -> int:  # noqa: N802
        ...

    @overload
    def addVertex(  # noqa: N802
        self, mapPoint: QgsPointXY, match: QgsPointLocator.Match  # noqa: N803
    ) -> int:
        ...

    def addVertex(self, *args, **kwargs) -> int:  # noqa: N802
        result = super().addVertex(*args, **kwargs)
        self.start_point_indicator_rubber_band.hide()
        return result

    @log_if_fails
    def canvasReleaseEvent(self, mouse_event: QgsMapMouseEvent) -> None:  # noqa: N802
        if self._tool_mode == ToolMode.PICK_SEGMENT:
            if mouse_event.button() == Qt.LeftButton:
                self._handle_pick_segment_left_click(mouse_event.mapPoint())
            return
        super().canvasReleaseEvent(mouse_event)

    def cadCanvasReleaseEvent(  # noqa: N802
        self, mouse_event: QgsMapMouseEvent
    ) -> None:
        """Override of the QgsMapToolAdvancedDigitizing.cadCanvasReleaseEvent

        This will receive adapted events from the cad system whenever a
        canvasReleaseEvent is triggered and it's not hidden by the cad's
        construction mode.

        Args:
            mouse_event (QgsMapMouseEvent): Mouse events prepared by the cad system
        """

        if self._tool_mode == ToolMode.RESHAPE:
            if mouse_event.button() == Qt.LeftButton:
                result = self.addVertex(
                    mouse_event.mapPoint(), mouse_event.mapPointMatch()
                )
                if result == AddVertexReturn.TransformationError:
                    MsgBar.warning(
                        tr("Cannot transform the point to the layers coordinate system")
                    )
                    return
                self.startCapturing()
            elif mouse_event.button() == Qt.RightButton:
                self._handle_reshape_right_click()

    def cadCanvasMoveEvent(self, mouse_event: QgsMapMouseEvent) -> None:  # noqa: N802
        if self._tool_mode == ToolMode.RESHAPE and self.size() == 0:
            self.start_point_indicator_rubber_band.movePoint(mouse_event.mapPoint())
        return super().cadCanvasMoveEvent(mouse_event)

    def _handle_pick_segment_left_click(self, location: QgsPointXY) -> None:
        with override_cursor(Qt.WaitCursor):
            common_segment, layer = self._find_common_segment(location)
            # No active layer or active layer feature found
            if layer is None:
                return

            if common_segment is None:
                MsgBar.info(
                    tr("No common segment found at location"),
                    tr(
                        "Features are not topologically connected "
                        "or a single vertex was clicked"
                    ),
                )
                return

            MsgBar.info(
                tr("Common segment found, changing to reshape mode"),
                success=True,
            )
            self._change_to_reshape_mode_for_geom(QgsGeometry(common_segment), layer)

    def _handle_reshape_right_click(self) -> None:
        with override_cursor(Qt.WaitCursor):
            new_geometry = self.captureCurve().curveToLine()
            self.stopCapturing()

            if new_geometry.isEmpty():
                self._change_to_pick_location_mode()
                return

            new_geometry.addZValue(self.defaultZValue())
            reshape_geom: Union[QgsPoint, QgsLineString] = new_geometry
            if new_geometry.numPoints() == 1:
                reshape_geom = new_geometry.pointN(0)

            reshape.make_reshape_edits(
                self.find_segment_results.common_parts,
                self.find_segment_results.edges,
                reshape_geom,
            )

            self._change_to_pick_location_mode()
            self.canvas().refresh()

    def _find_common_segment(
        self, location: QgsPointXY
    ) -> Tuple[Optional[QgsLineString], Optional[QgsVectorLayer]]:
        LOGGER.info("Calculating common segment")

        active_layer = iface.activeLayer()
        if active_layer is None:
            MsgBar.warning(tr("No active layer found"), tr("Activate a layer first"))
            return None, None

        results = SegmentReshapeTool.find_common_segment_at_location(
            location, self._identify_tool
        )

        if results is None:
            MsgBar.warning(
                tr("Did not find any active layer feature at mouse location")
            )
            return None, None

        self.find_segment_results = results

        return self.find_segment_results.segment, active_layer

    @staticmethod
    def find_common_segment_at_location(
        location: QgsPointXY, identify_tool: Optional[QgsMapToolIdentify] = None
    ) -> Optional[find_related.CommonGeometriesResult]:
        with _optional_identify_tool(identify_tool) as tool:
            identify_results = tool.identify(
                geometry=QgsGeometry.fromPointXY(location),
                mode=QgsMapToolIdentify.IdentifyMode.ActiveLayer,
                layerType=QgsMapToolIdentify.Type.VectorLayer,
            )

        if len(identify_results) < 1:
            return None

        identify_result = identify_results[0]
        layer, feature = (
            cast(QgsVectorLayer, identify_result.mLayer),
            identify_result.mFeature,
        )

        (
            *_,
            next_vertex_index,
            _,
        ) = feature.geometry().closestSegmentWithContext(location)

        return find_related.find_segment_to_reshape(
            layer, feature, (next_vertex_index - 1, next_vertex_index)
        )



#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of segment-reshape-qgis-plugin.
#
#  segment-reshape-qgis-plugin is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  segment-reshape-qgis-plugin is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with segment-reshape-qgis-plugin. If not, see <https://www.gnu.org/licenses/>.

import logging
from typing import Optional, cast

import qgis_plugin_tools
from qgis.gui import QgisInterface
from qgis.PyQt.QtCore import QCoreApplication, QObject, QTranslator
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction, QToolBar
from qgis.utils import iface as iface_
from qgis_plugin_tools.tools.custom_logging import setup_loggers
from qgis_plugin_tools.tools.i18n import setup_translation, tr
from qgis_plugin_tools.tools.resources import resources_path

import segment_reshape
import segment_reshape_plugin
from segment_reshape.map_tool.segment_reshape_tool import (
    SegmentReshapeTool,
    SegmentReshapeToolHandler,
)

iface = cast(QgisInterface, iface_)
LOGGER = logging.getLogger(__name__)


class SegmentReshapePlugin(QObject):
    def __init__(self) -> None:
        super().__init__(parent=None)

        self._teardown_loggers = lambda: None

        # Initialize locale
        locale, file_path = setup_translation()
        if file_path:
            self.translator = QTranslator()
            self.translator.load(file_path)
            QCoreApplication.installTranslator(self.translator)

        self.toolbar: Optional[QToolBar] = None
        self.segment_reshape_tool = SegmentReshapeTool(iface.mapCanvas())
        self.segment_reshape_tool_action: Optional[QAction] = None

    def initGui(self) -> None:  # noqa: N802 (qgis naming)
        self._teardown_loggers = setup_loggers(
            segment_reshape.__name__,
            segment_reshape_plugin.__name__,
            qgis_plugin_tools.__name__,
            message_log_name=tr("Segment reshape tool"),
        )

        toolbar = iface.addToolBar(
            self.tr("Segment reshape toolbar"),
        )
        toolbar.setObjectName("segment-reshape-toolbar")

        self.segment_reshape_tool_action = QAction(
            QIcon(resources_path("icons/segment_reshape.svg")),
            self.tr("Reshape common segment"),
            iface.mainWindow(),
        )
        self.segment_reshape_tool_handler = SegmentReshapeToolHandler(
            self.segment_reshape_tool, self.segment_reshape_tool_action
        )
        iface.registerMapToolHandler(self.segment_reshape_tool_handler)

        toolbar.addAction(self.segment_reshape_tool_action)

        self.toolbar = toolbar

    def unload(self) -> None:
        iface.unregisterMapToolHandler(self.segment_reshape_tool_handler)

        if self.toolbar is not None:
            self.toolbar.deleteLater()
        self.toolbar = None

        self._teardown_loggers()
        self._teardown_loggers = lambda: None



#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of segment-reshape-qgis-plugin.
#
#  segment-reshape-qgis-plugin is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  segment-reshape-qgis-plugin is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with segment-reshape-qgis-plugin. If not, see <https://www.gnu.org/licenses/>.

from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from typing import Dict, Iterator, List, Set, Tuple, Union

from qgis.core import (
    QgsFeature,
    QgsGeometry,
    QgsLineString,
    QgsPoint,
    QgsVectorLayer,
    QgsVertexId,
    QgsWkbTypes,
)

from segment_reshape.utils import clone_geometry_safely


class GeometryTransformationError(Exception):
    pass


@dataclass
class ReshapeCommonPart:
    layer: QgsVectorLayer
    feature: QgsFeature
    vertex_indices: List[int]
    is_reversed: bool
    """
    Indicates if this feature's geometry is digitized
    in reverse direction to the shared segment.
    """


@dataclass
class ReshapeEdge:
    layer: QgsVectorLayer
    feature: QgsFeature
    vertex_index: int
    is_start: bool
    """
    Indicates if this feature's vertex is at the
    start (or end) of the shared segment.
    """


_current_edit_command_layers: ContextVar[List[QgsVectorLayer]] = ContextVar(
    "current_edit_command_layers"
)
_set_editable_layer_ids: ContextVar[Set[str]] = ContextVar("set_editable_layer_ids")


@contextmanager
def _wrap_all_edit_commands() -> Iterator[None]:
    layers: List[QgsVectorLayer] = []
    set_editable_ids: Set[str] = set()
    token = _current_edit_command_layers.set(layers)
    editable_ids_token = _set_editable_layer_ids.set(set_editable_ids)
    try:
        yield
        for layer in layers:
            layer.endEditCommand()
    except GeometryTransformationError as e:
        for layer in layers:
            layer.destroyEditCommand()
            if layer.id() in set_editable_ids:
                layer.rollBack()
        raise e
    finally:
        _current_edit_command_layers.reset(token)
        _set_editable_layer_ids.reset(editable_ids_token)


def _set_editable_and_begin_edit_command_once(layer: QgsVectorLayer) -> None:
    if layer.isEditCommandActive():
        return

    if not layer.isEditable():
        if not layer.startEditing():
            raise GeometryTransformationError(f"could not start editing on {layer}")
        _set_editable_layer_ids.get().add(layer.id())

    layer.beginEditCommand("Reshape segment")
    _current_edit_command_layers.get().append(layer)


def make_reshape_edits(
    common_parts: List[ReshapeCommonPart],
    edges: List[ReshapeEdge],
    reshape_geometry: Union[QgsPoint, QgsLineString],
) -> None:
    """
    Reshape the provided common parts and edges, so that common part shared
    vertex indices are replaced by `reshape_geometry` and edges are moved
    to match `reshape_geometry`.

    Edits are made in a single edit command for each layer, and layers are
    set editable automatically if necessary.

    Raises `GeometryTransformationError` if operation fails.
    """

    # no crs support yet

    with _wrap_all_edit_commands():
        _reshape_common_parts(common_parts, reshape_geometry)
        if isinstance(reshape_geometry, QgsPoint):
            _move_edges(edges, reshape_geometry, reshape_geometry)
        else:
            _move_edges(
                edges,
                reshape_geometry.startPoint(),
                reshape_geometry.endPoint(),
            )


def _reshape_common_parts(
    common_parts: List[ReshapeCommonPart],
    reshape_geometry: Union[QgsPoint, QgsLineString],
) -> None:
    for common_part in common_parts:
        _set_editable_and_begin_edit_command_once(common_part.layer)
        new_geometry = _reshape_geometry(
            common_part.feature.geometry(),
            common_part.vertex_indices,
            (
                reshape_geometry.reversed()
                if isinstance(reshape_geometry, QgsLineString)
                and common_part.is_reversed
                else reshape_geometry
            ),
        )
        _update_geometry_to_layer_feature(
            common_part.layer,
            common_part.feature,
            new_geometry,
        )


def _reshape_geometry(
    original: QgsGeometry,
    vertex_indices: List[int],
    reshape_geometry: Union[QgsPoint, QgsLineString],
) -> QgsGeometry:
    new = clone_geometry_safely(original)

    if isinstance(reshape_geometry, QgsPoint):
        # move the first replaced vertex
        if not new.moveVertex(reshape_geometry, vertex_indices[0]):
            raise GeometryTransformationError(
                f"could not move vertex {vertex_indices[0]}"
                f" on {new} to {reshape_geometry}"
            )
        # delete rest in reverse order
        for deleted_index in sorted(vertex_indices[1:], reverse=True):
            if not new.deleteVertex(deleted_index):
                raise GeometryTransformationError(
                    f"could not delete vertex {deleted_index} on {new}"
                )

    else:
        # since the indices are always the longest continuous segment on
        # the origin geometry, if the indices wrap around also at the index
        # list start/end, its known to be a closed geometry

        # handle the case when a full polygon ring is reshaped
        if (
            original.type() == QgsWkbTypes.GeometryType.PolygonGeometry
            and len(vertex_indices) > 1
            and vertex_indices[0] == vertex_indices[-1]
        ):
            # can just omit the last vertex from target indices since it will
            # either be a mid-ring index duplicated, or the polygon ring origin index
            # (first/last, which will be automatically matched even without explicitly
            # moving both indices)
            vertex_indices = vertex_indices[:-1]

            # support reshape both with a fully redrawn closed geometry, and even
            # without explicitly closing the geometry, similary to target indices
            # can just omit the last reshape geometry vertex if it was closed
            if reshape_geometry.isClosed():
                reshape_geometry = reshape_geometry.clone()
                if not reshape_geometry.deleteVertex(
                    QgsVertexId(0, 0, reshape_geometry.vertexCount(0, 0) - 1)
                ):
                    raise GeometryTransformationError(
                        f"could not delete last vertex on {reshape_geometry}"
                    )

        # handle the case when a full closed linestring in reshaped
        elif (
            original.type() == QgsWkbTypes.GeometryType.LineGeometry
            and len(vertex_indices) > 1
            and vertex_indices[0] == vertex_indices[-1]
        ):
            # handle wraparound at origin by simply using the correct start index
            # instead of duplicating the max vertex index at the first index
            if vertex_indices[0] == max(vertex_indices):
                vertex_indices[0] = min(vertex_indices) - 1

            # handle wraparound at non-origin by simply rewriting the indices as if
            # the origin was at the new reshape geometry origin. this will essentially
            # scroll origin along the original geometryto the reshape start location,
            # assume this is not an issue since there was nothing connected at the
            # previous origin and something was connected at the new origin
            else:
                vertex_indices = list(
                    range(min(vertex_indices) - 1, max(vertex_indices) + 1)
                )

            # similarly to polygon rings support the reshape even without closing the
            # reshape geometry, by closing the reshape geometry manually here
            # NOTE: this way the code will never break closed linestring geometries,
            # even if that is what is wanted. TODO: possibly change this logic to
            # always require closed reshape geometries for closed origin geometries,
            # so that its an error to reshape a polygon ring without closing it?
            if not reshape_geometry.isClosed():
                reshape_geometry = reshape_geometry.clone()
                reshape_geometry.close()

        # handle case when a closed linestring in partially reshaped in a way that
        # the part origin wraparound falls inside the target vertex indices as
        # indicated by a gap in the otherwise continuous vertex indices
        elif (
            original.type() == QgsWkbTypes.GeometryType.LineGeometry
            and len(vertex_indices) > 1
            and any(
                abs(first - second) > 1
                for first, second in zip(vertex_indices[:-1], vertex_indices[1:])
            )
        ):
            # reshape will scroll the origin to be located at the start
            # of the reshape segment, to match the ends at the new position make
            # the wraparound vertex of the list the part minimum instead of the
            # part maximum (to insert the reshape as the start of the part), and
            # move the part maximum index to the start of the reshape to match
            # the new origin
            min_index, max_index = min(vertex_indices), max(vertex_indices)
            vertex_indices = [
                (min_index - 1 if i == max_index else i) for i in vertex_indices
            ]
            if not new.moveVertex(reshape_geometry.startPoint(), max_index):
                raise GeometryTransformationError(
                    f"could not move vertex {max_index}"
                    f" on {new} to {reshape_geometry.startPoint()}"
                )

        # add vertices before the first replaced vertex
        min_vertex_index = min(vertex_indices)
        for new_vertex_index, new_vertex in enumerate(reshape_geometry.vertices()):
            add_before_index = min_vertex_index + new_vertex_index
            if not new.insertVertex(new_vertex, add_before_index):
                raise GeometryTransformationError(
                    f"could not insert {new_vertex} vertex"
                    f" before {add_before_index} on {new}"
                )

        # delete replace vertices that were moved by the added count in reverse order
        added_vertex_count = reshape_geometry.vertexCount()
        for original_delete_index in sorted(vertex_indices, reverse=True):
            deleted_index = original_delete_index + added_vertex_count
            if not new.deleteVertex(deleted_index):
                raise GeometryTransformationError(
                    f"could not delete vertex {deleted_index} on {new}"
                )

    return new


def _move_edges(
    edges: List[ReshapeEdge], new_start: QgsPoint, new_end: QgsPoint
) -> None:
    # hold on to the updates so next iteration for same feature
    # uses the previous updated geometry instead of initial geometry
    previously_updated_geoms: Dict[Tuple[str, int], QgsGeometry] = {}

    for edge in edges:
        _set_editable_and_begin_edit_command_once(edge.layer)
        new_geometry = _move_vertex(
            previously_updated_geoms.get(
                (edge.layer.id(), edge.feature.id()), edge.feature.geometry()
            ),
            edge.vertex_index,
            new_start if edge.is_start else new_end,
        )
        _update_geometry_to_layer_feature(
            edge.layer,
            edge.feature,
            new_geometry,
        )
        previously_updated_geoms[(edge.layer.id(), edge.feature.id())] = new_geometry


def _move_vertex(
    original: QgsGeometry, vertex_index: int, new_position: QgsPoint
) -> QgsGeometry:
    new = clone_geometry_safely(original)
    if not new.moveVertex(new_position, vertex_index):
        raise GeometryTransformationError(
            f"could not move vertex {vertex_index} on {new} to {new_position}"
        )
    return new


def _update_geometry_to_layer_feature(
    layer: QgsVectorLayer,
    feature: QgsFeature,
    new_geometry: QgsGeometry,
) -> None:
    if not layer.changeGeometry(
        feature.id(),
        new_geometry,
    ):
        raise GeometryTransformationError(
            f"could not update geometry {new_geometry} to fid {feature.id()} on {layer}"
        )
